@@ -1,5 +1,5 @@
 # -*- coding: utf-8 -*-
-# Copyright (C) 2008-2016, Luis Pedro Coelho <luis@luispedro.org>
+# Copyright (C) 2008-2017, Luis Pedro Coelho <luis@luispedro.org>
 # vim: set ts=4 sts=4 sw=4 expandtab smartindent:
 # LICENSE: MIT
 '''
@@ -50,6 +50,19 @@ class TaskletMixin(object):
     def __getitem__(self, slice):
         return Tasklet(self, _getitem(slice))
 
+    def __iter__(self):
+        raise NotImplementedError("""\
+You cannot iterate over Task results.
+
+If the Task returns an iterator of a fixed size, `iteratetask` may fit your usage:
+
+http://jug.readthedocs.io/en/latest/api.html?highlight=iteratetask#jug.iteratetask
+
+Otherwise, you may wish to consider `bvalue`:
+
+http://jug.readthedocs.io/en/latest/api.html?highlight=bvalue#jug.bvalue
+""")
+
 class Task(TaskletMixin):
     '''
     T = Task(f, dep0, dep1,..., kw_arg0=kw_val0, kw_arg1=kw_val1, ...)
@@ -618,6 +631,44 @@ class iteratetask(object):
     def __len__(self):
         return self.n
 
+def _get_check(r, i, n):
+    if len(r) != n:
+        raise ValueError("Expected a tuple of size {} got {}".format(n, len(r)))
+    return r[i]
+
+def return_tuple(n):
+    '''
+    Wraps a TaskGenerator which returns a tuple of size `n`. Now, the
+    TaskGenerator will also return a tuple, allowing for more natural code.
+
+    Examples
+    --------
+
+    ::
+
+        @return_tuple(2)
+        @TaskGenerator
+        def plus1(x):
+            return x, 1+x
+
+        x,x1 = plus1(0)
+
+
+
+
+
+    Parameters
+    ----------
+    n : int
+        Expected size of tuple (this is checked at runtime)
+    '''
+    from functools import partial
+    def wrapper(f):
+        def wrapped(*args, **kwargs):
+            r = f(*args, **kwargs)
+            return tuple([Tasklet(r, partial(_get_check, i=i, n=n)) for i in range(n)])
+        return wrapped
+    return wrapper
 
 def describe(t):
     '''
