@@ -1,5 +1,5 @@
 # -*- coding: utf-8 -*-
-# Copyright (C) 2009-2016, Luis Pedro Coelho <luis@luispedro.org>
+# Copyright (C) 2009-2017, Luis Pedro Coelho <luis@luispedro.org>
 # vim: set ts=4 sts=4 sw=4 expandtab smartindent:
 #
 # Permission is hereby granted, free of charge, to any person obtaining a copy
@@ -22,13 +22,17 @@
 
 
 import os
+import re
+from functools import partial
 
 from .task import Task, TaskGenerator, Tasklet, value
 
 __all__ = [
     'timed_path',
     'identity',
+    'jug_execute',
     'CustomHash',
+    'sync_move',
     ]
 
 def _identity(x):
@@ -77,8 +81,8 @@ class CustomHash(object):
     This is an advanced feature and you can shoot yourself in the foot with it.
     Make sure you know what you are doing. In particular, hash_function should
     be a strong hash: ``hash_function(obj0) == hash_function(obj1)`` is taken
-    to imply that ``obj0 == obj1``
-
+    to imply that ``obj0 == obj1``. The hash function should return a ``bytes``
+    object.
 
     You can use the helpers in the ``jug.hash`` module (in particular
     ``hash_one``) to help you. The implementation of ``timed_path`` is a good
@@ -154,12 +158,32 @@ def timed_path(path):
     '''
     return CustomHash(path, hash_with_mtime_size)
 
+
+def prepare_task_matcher(pattern):
+    if re.match(r'/.*/', pattern):
+        # Looks like a regular expression
+        regex = re.compile(pattern.strip('/'))
+
+    elif '.' in pattern:
+        # Looks like a full task name
+        regex = pattern.replace('.', '\\.')
+
+    else:
+        # A bare function name perhaps?
+        regex = re.compile(r'\.' + pattern)
+
+    return partial(re.search, regex)
+
+
 @TaskGenerator
-def jug_execute(args, check_exit=True, run_after=None):
-    '''jug_execute(args, check_exit=True, run_after=None)
+def jug_execute(args, check_exit=True, run_after=None, return_value=None):
+    '''jug_execute(args, check_exit=True, run_after=None, return_value=None)
 
     Wrapper around ``subprocess.call()``
 
+    The ``run_after`` and ``return_value`` arguments are used only for setting
+    dependencies.
+
     Examples
     --------
 
@@ -168,6 +192,13 @@ def jug_execute(args, check_exit=True, run_after=None):
         create_input_tmp = jug_execute(['cp', 'input', 'input.tmp'])
         jug_execute(['wc', '-l', 'input.tmp'], run_after=create_input_tmp)
 
+        # Use the return_value argument:
+
+        oname = 'input.tmp'
+        oname = jug_execute(['cp', 'input', oname], return_value=oname)
+        jug_execute(['wc', '-l', oname]) # <- Now depends on previous task
+
+
     Parameters
     ----------
     args : list of str
@@ -176,8 +207,28 @@ def jug_execute(args, check_exit=True, run_after=None):
     run_after : any, optional
         This is unused by the function, but can be used to order different
         calls to jug_execute
+    return_value : any, optional
+        A value to return. Used to induce task dependencies
     '''
     import subprocess
     ret = subprocess.call(args)
     if check_exit and ret != 0:
         raise SystemError("Error in system")
+    return return_value
+
+def sync_move(src, dst):
+    '''Sync the file and move it
+
+    This ensures that the move is truly atomic
+
+    Parameters
+    ----------
+    src : filename
+        Source file
+    dst: filename
+        Destination file
+    '''
+    from jug.backends.file_store import fsync_dir
+    from os import rename
+    fsync_dir(src)
+    rename(src, dst)
